import MetalKit
import simd

extension float4x4 {
    
    init(orthoLeft left: Float, right: Float, bottom: Float, top: Float, near: Float, far: Float) {
        let sx = 2.0 / (right - left)
        let sy = 2.0 / (top - bottom)
        let sz = -1.0 / (far - near)
        let tx = -(right + left) / (right - left)
        let ty = -(top + bottom) / (top - bottom)
        let tz = near * sz
        self.init(
            float4( sx, 0.0, 0.0, 0.0),
            float4(0.0,  sy, 0.0, 0.0),
            float4(0.0, 0.0,  sz, 0.0),
            float4( tx,  ty,  tz, 1.0)
        )
    }
    
    init(translationX tx: Float, y ty: Float, z tz: Float) {
        self.init(
            float4(1.0, 0.0, 0.0, 0.0),
            float4(0.0, 1.0, 0.0, 0.0),
            float4(0.0, 0.0, 1.0, 0.0),
            float4( tx,  ty,  tz, 1.0)
        )
    }
}

class Renderer : NSObject, MTKViewDelegate {
    
    let view: MTKView
    let device: MTLDevice!
    let commandQueue: MTLCommandQueue!
    
    var vertexBuffer: MTLBuffer!
    var positionBuffer: MTLBuffer!
    var colorBuffer: MTLBuffer!
    
    var interleavedRenderPipeline: MTLRenderPipelineState!
    var noninterleavedRenderPipeline: MTLRenderPipelineState!
    
    init(_ view: MTKView) {
        self.view = view
        device = view.device
        commandQueue = device.makeCommandQueue()
        super.init()
        
        view.delegate = self
        makeResources()
        makePipelineStates()
    }
    
    func makeResources() {
        let vertices = UnsafeMutableBufferPointer<ColoredVertex>.allocate(capacity: 3)
        defer { vertices.deallocate() }
        vertices[0].position = float3(0.0, -100.0, 0)
        vertices[0].color = float3(1.0, 0.0, 0.0)
        vertices[1].position = float3(-100.0, 100.0, 0)
        vertices[1].color = float3(0.0, 1.0, 0.0)
        vertices[2].position = float3(100.0, 100.0, 0)
        vertices[2].color = float3(0.0, 0.0, 1.0)
        vertexBuffer = device.makeBuffer(bytes: vertices.baseAddress!,
                                         length: MemoryLayout<ColoredVertex>.stride * vertices.count,
                                         options: [.storageModeShared])
        
        let positions = UnsafeMutableBufferPointer<float3>.allocate(capacity: 3)
        defer { positions.deallocate() }
        positions[0] = vertices[0].position
        positions[1] = vertices[1].position
        positions[2] = vertices[2].position
        positionBuffer = device.makeBuffer(bytes: positions.baseAddress!,
                                           length: MemoryLayout<float3>.stride * positions.count,
                                           options: [.storageModeShared])
        
        let colors = UnsafeMutableBufferPointer<float3>.allocate(capacity: 3)
        defer { colors.deallocate() }
        colors[0] = vertices[0].color
        colors[1] = vertices[1].color
        colors[2] = vertices[2].color
        colorBuffer = device.makeBuffer(bytes: colors.baseAddress!,
                                        length: MemoryLayout<float3>.stride * colors.count,
                                        options: [.storageModeShared])
    }
    
    func makePipelineStates() {
        let library = device.makeDefaultLibrary()!
        
        let renderPipelineDescriptor = MTLRenderPipelineDescriptor()
        renderPipelineDescriptor.colorAttachments[0].pixelFormat = view.colorPixelFormat
        renderPipelineDescriptor.vertexFunction = library.makeFunction(name: "vertex_main")!
        renderPipelineDescriptor.fragmentFunction = library.makeFunction(name: "fragment_main")!
        
        let interleavedVertexDescriptor = MTLVertexDescriptor()
        interleavedVertexDescriptor.attributes[0].bufferIndex = 0
        interleavedVertexDescriptor.attributes[0].offset = 0
        interleavedVertexDescriptor.attributes[0].format = .float3
        interleavedVertexDescriptor.attributes[1].bufferIndex = 0
        interleavedVertexDescriptor.attributes[1].offset = MemoryLayout<float3>.stride
        interleavedVertexDescriptor.attributes[1].format = .float3
        interleavedVertexDescriptor.layouts[0].stride = MemoryLayout<float3>.stride * 2
        renderPipelineDescriptor.vertexDescriptor = interleavedVertexDescriptor
        interleavedRenderPipeline = try! device.makeRenderPipelineState(descriptor: renderPipelineDescriptor)
        
        let noninterleavedVertexDescriptor = MTLVertexDescriptor()
        noninterleavedVertexDescriptor.attributes[0].bufferIndex = 0
        noninterleavedVertexDescriptor.attributes[0].offset = 0
        noninterleavedVertexDescriptor.attributes[0].format = .float3
        noninterleavedVertexDescriptor.attributes[1].bufferIndex = 1
        noninterleavedVertexDescriptor.attributes[1].offset = 0
        noninterleavedVertexDescriptor.attributes[1].format = .float3
        noninterleavedVertexDescriptor.layouts[0].stride = MemoryLayout<float3>.stride
        noninterleavedVertexDescriptor.layouts[1].stride = MemoryLayout<float3>.stride
        renderPipelineDescriptor.vertexDescriptor = noninterleavedVertexDescriptor
        noninterleavedRenderPipeline = try! device.makeRenderPipelineState(descriptor: renderPipelineDescriptor)
    }
    
    func mtkView(_ view: MTKView, drawableSizeWillChange size: CGSize) {
    }
    
    func draw(in view: MTKView) {
        guard let renderPassDescriptor = view.currentRenderPassDescriptor else { return }
        guard let commandBuffer = commandQueue.makeCommandBuffer() else { return }
        
        let renderCommandEncoder = commandBuffer.makeRenderCommandEncoder(descriptor: renderPassDescriptor)!
        
        let halfWidth = Float(view.bounds.width * 0.5)
        let halfHeight = Float(view.bounds.height * 0.5)
        
        let projectionMatrix = float4x4(orthoLeft: -halfWidth,
                                        right: halfWidth,
                                        bottom: halfHeight,
                                        top: -halfHeight,
                                        near: -1.0,
                                        far: 1.0)
        var leftMatrix = projectionMatrix * float4x4(translationX: -200, y: 0, z: 0)
        var rightMatrix = projectionMatrix * float4x4(translationX: 200, y: 0, z: 0)
        
        renderCommandEncoder.setVertexBuffer(vertexBuffer, offset: 0, index: 0)
        renderCommandEncoder.setVertexBytes(&leftMatrix, length: MemoryLayout<float4x4>.stride, index: 2)
        renderCommandEncoder.setRenderPipelineState(interleavedRenderPipeline)
        renderCommandEncoder.drawPrimitives(type: .triangle, vertexStart: 0, vertexCount: 3)
        
        renderCommandEncoder.setVertexBuffer(positionBuffer, offset: 0, index: 0)
        renderCommandEncoder.setVertexBuffer(colorBuffer, offset: 0, index: 1)
        renderCommandEncoder.setVertexBytes(&rightMatrix, length: MemoryLayout<float4x4>.stride, index: 2)
        renderCommandEncoder.setRenderPipelineState(noninterleavedRenderPipeline)
        renderCommandEncoder.drawPrimitives(type: .triangle, vertexStart: 0, vertexCount: 3)
        
        renderCommandEncoder.endEncoding()
        
        if let drawable = view.currentDrawable {
            commandBuffer.present(drawable)
        }
        
        commandBuffer.commit()
    }
}
